package meerkat;

import com.google.protobuf.ByteString;
import com.google.protobuf.InvalidProtocolBufferException;
import meerkat.ProtobufsMessages.Tag;
import meerkat.bulletinboard.SimpleBulletinBoardClient;
import meerkat.comm.CommunicationException;
import meerkat.crypto.Encryption;
import meerkat.protobuf.BulletinBoardAPI;
import meerkat.protobuf.Crypto;
import util.AccurateTimestamp;
import util.CollectionMessagesUtils;
import util.RegistryTags;

import java.io.IOException;
import java.text.ParseException;
import java.util.*;

/**
 * Created by Vladimir Eliezer Tokarev on 1/8/2016.
 * Gives the ability to manage voters information
 */
public class SimpleRegistry {

    protected Encryption signatory;

    protected SimpleBulletinBoardClient communicator;

    public SimpleRegistry(Encryption signatory, SimpleBulletinBoardClient communicator) {
        this.signatory = signatory;
        this.communicator = communicator;
    }

    /**
     * Creates BulletinBoardMessage with signed basicMessage and UnsignedBulletinBoardMessage that contains the basic message
     * @param basicMessage BasicMessage
     * @return BulletinBoardAPI.BulletinBoardMessage
     */
    private BulletinBoardAPI.BulletinBoardMessage CreateBulletinBoardMessage(BulletinBoardAPI.UnsignedBulletinBoardMessage basicMessage) throws IOException {
        BulletinBoardAPI.BulletinBoardMessage.Builder bulletinBoardMessage =
                BulletinBoardAPI.BulletinBoardMessage.newBuilder();

        Crypto.RerandomizableEncryptedMessage encryptedMessage =
                signatory.encrypt(basicMessage, signatory.generateRandomness(new Random()));

        Crypto.Signature.Builder messageSignature = Crypto.Signature.newBuilder();
        messageSignature.mergeFrom(encryptedMessage);

        bulletinBoardMessage.setMsg(basicMessage);

        bulletinBoardMessage.addSig(messageSignature);

        return bulletinBoardMessage.build();
    }

    /**
     * Gets messages that have the wanted id tag and have or the ADD_TO_GROUP_TAG or REMOVE_FROM_GROUP_TAG
     *
     * @param tags the tags based on which the messages will be filtered
     * @return List<BulletinBoardAPI.BulletinBoardMessage>
     */
    private List<BulletinBoardAPI.BulletinBoardMessage> GetRelevantMessages(List<String> tags) {
        BulletinBoardAPI.MessageFilterList.Builder filters =
                BulletinBoardAPI.MessageFilterList.newBuilder();

        for (String tag : tags) {
            BulletinBoardAPI.MessageFilter.Builder idFilter =
                    BulletinBoardAPI.MessageFilter.newBuilder().setTag(tag);

            filters.addFilter(idFilter);
        }
        return communicator.readMessages(filters.build());
    }

    /**
     * Gets Relevant bulletinBoard messages from communicator than converts them to VoterRegistryMessages
     *
     * @param tags list of tags that will be used as filters
     * @return List<VoterRegistryMessage>
     */
    private List<VoterRegistryMessage> GetRelevantVoterRegistryMessages(List<String> tags) throws InvalidProtocolBufferException {
        List<BulletinBoardAPI.BulletinBoardMessage> relevantMessages = GetRelevantMessages(tags);
        List<BulletinBoardAPI.UnsignedBulletinBoardMessage> messages =
                CollectionMessagesUtils.GetUnsignedBulletinBoardMessages(relevantMessages);
        return CollectionMessagesUtils.ConvertToVoterRegistryMessages(messages);
    }


    /**
     * Adds new voter to the bulletin-board
     * @param voterID id of the new user
     * @param personalData for example residence location
     * @return void
     * @throws CommunicationException, IOException
     */
    public void AddVoter(String voterID, String personalData) throws CommunicationException, IOException {
        BulletinBoardAPI.UnsignedBulletinBoardMessage.Builder basicMessage =
                BulletinBoardAPI.UnsignedBulletinBoardMessage.newBuilder().
                        addTag(RegistryTags.ID_TAG + " " + voterID)
                        .addTag(RegistryTags.VOTER_ENTRY_TAG.toString())
                        .addTag(RegistryTags.ACTION_TIMESTAMP_TAG + " " + AccurateTimestamp.GetCurrentTimestampString());

        basicMessage.setData(ByteString.copyFrom(personalData.getBytes()));

        communicator.postMessage(CreateBulletinBoardMessage(basicMessage.build()));
    }

    /**
     * Adding given voter to given group
     *
     * @param voterID the voter id
     * @param groupID the group id to which we want add the voter
     * @return true if the adding action succeeded else return false
     * @throws CommunicationException
     */
    public void AddToGroup(String voterID, String groupID) throws CommunicationException, IOException {
        BulletinBoardAPI.UnsignedBulletinBoardMessage.Builder basicMessage =
                BulletinBoardAPI.UnsignedBulletinBoardMessage.newBuilder()
                        .addTag(RegistryTags.ID_TAG + " " + voterID)
                        .addTag(RegistryTags.GROUP_ID_TAG + " " + groupID)
                        .addTag(RegistryTags.GROUP_ACTION_TAG + " " + RegistryTags.ADD_TO_GROUP_TAG)
                        .addTag(RegistryTags.ACTION_TIMESTAMP_TAG + " " + AccurateTimestamp.GetCurrentTimestampString());

        communicator.postMessage(CreateBulletinBoardMessage(basicMessage.build()));
    }

    /**
     * Removes given voter from given group
     *
     * @param voterID the voter id
     * @param groupID the group from which we want to remove the user
     * @return true if the removing action succeeded else return false
     * @throws CommunicationException
     */
    public void RemoveFromGroup(String voterID, String groupID) throws CommunicationException, IOException {
        BulletinBoardAPI.UnsignedBulletinBoardMessage.Builder basicMessage =
                BulletinBoardAPI.UnsignedBulletinBoardMessage.newBuilder()
                        .addTag(RegistryTags.ID_TAG + " " + voterID)
                        .addTag(RegistryTags.GROUP_ID_TAG + " " + groupID)
                        .addTag(RegistryTags.GROUP_ACTION_TAG + " " + RegistryTags.REMOVE_FROM_GROUP_TAG)
                        .addTag(RegistryTags.ACTION_TIMESTAMP_TAG + " " + AccurateTimestamp.GetCurrentTimestampString());

        communicator.postMessage(CreateBulletinBoardMessage(basicMessage.build()));
    }

    /**
     * Sets that the voter have voted
     *
     * @param id id tag string
     * @return true if the set voted succeded else false
     * @throws CommunicationException
     */
    public void AddVoter(String id) throws CommunicationException, IOException {
        BulletinBoardAPI.UnsignedBulletinBoardMessage.Builder basicMessage =
                BulletinBoardAPI.UnsignedBulletinBoardMessage.newBuilder()
                        .addTag(RegistryTags.ID_TAG + " " + id)
                        .addTag(RegistryTags.VOTE_ACTION_TAG.toString())
                        .addTag(RegistryTags.ACTION_TIMESTAMP_TAG + " " + AccurateTimestamp.GetCurrentTimestampString());

        communicator.postMessage(CreateBulletinBoardMessage(basicMessage.build()));
    }


    /**
     * Requests all the groups that the given id voter is in
     *
     * @param id id tag string
     * @return list of groups ids (or names), if the method fails its empty
     * @throws CommunicationException, InvalidProtocolBufferException
     */
    public List<String> GetGroups(String id) throws CommunicationException, InvalidProtocolBufferException {
        List<String> GroupsActionsTags = new ArrayList<String>(){{
            add(RegistryTags.ID_TAG + " " + id);
            add(RegistryTags.GROUP_ACTION_TAG + " " + RegistryTags.REMOVE_FROM_GROUP_TAG);
            add(RegistryTags.GROUP_ACTION_TAG + " " + RegistryTags.ADD_TO_GROUP_TAG);
        }};

        List<VoterRegistryMessage> voterRegistryMessages = GetRelevantVoterRegistryMessages(GroupsActionsTags);

        try {
            Map<String, VoterRegistryMessage> groupIdToMessage = CollectionMessagesUtils.GetLatestGroupsActions(voterRegistryMessages);
            return CollectionMessagesUtils.GetListOfGroupIds(groupIdToMessage);
        } catch (ParseException e) {
            return null;
        }
    }

    /**
     * Retrieves list of strings that represents voter
     *
     * @param id id tag string
     * @return list of strings (empty list if the lookup failed)
     * @throws CommunicationException
     */
    public List<String> GetPersonIDDetails(String id) throws CommunicationException, InvalidProtocolBufferException, ParseException {
        List<Tag> GroupsActionsTags = new ArrayList<>();
        GroupsActionsTags.add(Tag.newBuilder().setContent(RegistryTags.ID_TAG + " " + id).build());
        GroupsActionsTags.add(Tag.newBuilder().setContent(RegistryTags.VOTER_ENTRY_TAG.toString()).build());

        List<VoterRegistryMessage> voterRegistryMessages = GetRelevantVoterRegistryMessages(GroupsActionsTags);
        VoterRegistryMessage LatestMessage = voterRegistryMessages.get(0);

        for (VoterRegistryMessage message : voterRegistryMessages) {
            if (message.GetBasicMessageActionTimestamp().before(LatestMessage.GetBasicMessageActionTimestamp())) {
                LatestMessage = message;
            }
        }

        return Arrays.asList(LatestMessage.GetWantedTagFromBasicMessage(RegistryTags.ID_TAG), LatestMessage.base.getData().toString());
    }
}
